from PySide6.QtCore import QObject, Signal as pyqtSignal

class Common(QObject):
    xpix = ypix = zpix = 0
    world_x = world_y = world_z = None
    world_x_str = world_y_str = world_z_str = ""
    xy_hline = xy_vline = xz_hline = xz_vline = zy_hline = zy_vline = None
    canvas_xy = canvas_xz = canvas_zy = None
    _background_xy = _background_xz = _background_zy = None
    im_xy = im_xz = im_zy = None
    ax_xy = ax_xz = ax_zy = None
    overlay_ax_xy = overlay_ax_xz = overlay_ax_zy = None
    plabel_xy = plabel_xz = plabel_zy = None
    chlabel_xy = chlabel_xz = chlabel_zy = None
    chval_box_xy = chval_box_xz = chval_box_zy = None 
    ax_xy_xunit = ax_xy_yunit = ax_xz_xunit = ax_xz_yunit = ax_zy_xunit = ax_zy_yunit = None
    ax_xy_xtype = ax_xy_ytype = ax_xz_xtype = ax_xz_ytype = ax_zy_xtype = ax_zy_ytype = None
    slider_xy = slider_xz = slider_zy = None
    clicked_xy = clicked_xz = clicked_zy = False
    colorbar_xy = colorbar_xz = colorbar_zy = None
    cax_xy = cax_xz = cax_zy = None
    fig_xy = fig_xz = fig_zy = None
    ax_coord_xy = ax_coord_xz = ax_coord_zy = None
    xmin_input_xy = xmax_input_xy = xmin_input_xz = xmax_input_xz = xmin_input_zy = xmax_input_zy = None
    ymin_input_xy = ymax_input_xy = ymin_input_xz = ymax_input_xz = ymin_input_zy = ymax_input_zy = None
    zmin_input_xy = zmax_input_xy = zmin_input_xz = zmax_input_xz = zmin_input_zy = zmax_input_zy = None
    integ_cax = []
    integ_colorbar = []
    ch_cax = None
    ch_colorbar = None
    position_updated = pyqtSignal(int, int, int)
    main_window = None
    
    @classmethod
    def update_pix(cls, x, y, z):
        cls.xpix, cls.ypix, cls.zpix = x, y, z
        try:
            from takefits.tools.spectrum import SpecWindow
            if SpecWindow.is_on:
                instance = cls._get_instance()
                instance.position_updated.emit(int(cls.xpix), int(cls.ypix), int(cls.zpix))
        except ImportError:
            pass

    @classmethod
    def _get_instance(cls):
        if not hasattr(cls, '_instance'):
            cls._instance = cls()
        return cls._instance
        
    @classmethod
    def update_lines(cls, plane, hline, vline):
        if plane == 'xy':
            cls.xy_hline = hline
            cls.xy_vline = vline
        elif plane == 'xz':
            cls.xz_hline = hline
            cls.xz_vline = vline
        elif plane == 'zy':
            cls.zy_hline = hline
            cls.zy_vline = vline

    @classmethod
    def update_background(cls, plane, background):
        if plane == 'xy':
            cls._background_xy = background
        elif plane == 'xz':
            cls._background_xz = background
        elif plane == 'zy':
            cls._background_zy = background

    @classmethod
    def copy_overlay_background(cls, plane):
        """Copies the overlay background, hiding regions in the main window when needed."""
        canvas = getattr(cls, f'canvas_{plane}', None)
        overlay_ax = getattr(cls, f'overlay_ax_{plane}', None)
        if canvas is None or overlay_ax is None:
            return None

        hidden_patches = []
        hidden_markers = []
        region_manager = None
        marker_manager = None
        viewer = None
        main = getattr(cls, 'main_window', None)
        if main is not None:
            if getattr(main, 'plane', 'xy') == plane or plane == 'xy':
                viewer = main
            else:
                for subwindow in getattr(main, 'subwindows', []):
                    if getattr(subwindow, 'plane', None) == plane:
                        viewer = subwindow
                        break
        if viewer is not None:
            region_manager = getattr(viewer, 'region_manager', None)
            marker_manager = getattr(viewer, 'marker_manager', None)
            if region_manager is not None:
                hidden_patches = region_manager.prepare_for_background_capture()
            if marker_manager is not None:
                hidden_markers = marker_manager.prepare_for_background_capture(plane)

        background = canvas.copy_from_bbox(overlay_ax.bbox)

        if region_manager is not None:
            region_manager.restore_after_background_capture(hidden_patches)
            region_manager.draw_regions_for_blit()
        if marker_manager is not None:
            if hidden_markers:
                marker_manager.restore_after_background_capture(hidden_markers)
            marker_manager.draw_markers_for_blit()
            if plane == 'xy':
                marker_manager.redraw_planes(['xz', 'zy'])

        return background

    @classmethod
    def update_im(cls, plane, im):
        if plane == 'xy': cls.im_xy = im
        elif plane == 'xz': cls.im_xz = im
        elif plane == 'zy': cls.im_zy = im
        
    @classmethod
    def update_ax(cls, plane, ax):
        if plane == 'xy': cls.ax_xy = ax
        elif plane == 'xz': cls.ax_xz = ax
        elif plane == 'zy': cls.ax_zy = ax

    @classmethod
    def update_overlay_ax(cls, plane, ax):
        if plane == 'xy': cls.overlay_ax_xy = ax
        elif plane == 'xz': cls.overlay_ax_xz = ax
        elif plane == 'zy': cls.overlay_ax_zy = ax

    @classmethod
    def update_canvas(cls, plane, canvas):
        if plane == 'xy': cls.canvas_xy = canvas
        elif plane == 'xz': cls.canvas_xz = canvas
        elif plane == 'zy': cls.canvas_zy = canvas

    @classmethod
    def update_poslabel(cls, plane, label):
        if plane == 'xy': cls.plabel_xy = label
        elif plane == 'xz': cls.plabel_xz = label
        elif plane == 'zy': cls.plabel_zy = label
        
    @classmethod
    def update_chlabel(cls, plane, label):
        if plane == 'xy': cls.chlabel_xy = label
        elif plane == 'xz': cls.chlabel_xz = label
        elif plane == 'zy': cls.chlabel_zy = label   

    @classmethod
    def update_hpbw(cls, plane, hpbw):
        if plane == 'xy': cls.hpbw_xy = hpbw
        elif plane == 'xz': cls.hpbw_xz = hpbw
        elif plane == 'zy': cls.hpbw_zy = hpbw
        
    @classmethod
    def update_slider(cls, plane, slider):
        if plane == 'xy': cls.slider_xy = slider
        elif plane == 'xz': cls.slider_xz = slider
        elif plane == 'zy': cls.slider_zy = slider       

    @classmethod
    def update_chval_box(cls, plane, textbox):
        if plane == 'xy': cls.chval_box_xy = textbox
        elif plane == 'xz': cls.chval_box_xz = textbox
        elif plane == 'zy': cls.chval_box_zy = textbox    
    
        
    @classmethod
    def update_ax_units(cls, plane, xunit, yunit):
        if plane == 'xy': cls.ax_xy_xunit, cls.ax_xy_yunit = xunit, yunit
        elif plane == 'xz': cls.ax_xz_xunit, cls.ax_xz_yunit = xunit, yunit
        elif plane == 'zy': cls.ax_zy_xunit, cls.ax_zy_yunit = xunit, yunit

    @classmethod
    def update_ax_types(cls, plane, xtype, ytype):
        if plane == 'xy': cls.ax_xy_xtype, cls.ax_xy_ytype = xtype, ytype
        elif plane == 'xz': cls.ax_xz_xtype, cls.ax_xz_ytype = xtype, ytype
        elif plane == 'zy': cls.ax_zy_xtype, cls.ax_zy_ytype = xtype, ytype

    @classmethod
    def update_world_xyz(cls, x, y, z):
        cls.world_x, cls.world_y, cls.world_z = x, y, z
    
    @classmethod
    def update_world_xyz_str(cls, x, y, z):
        cls.world_x_str, cls.world_y_str, cls.world_z_str = x, y, z
        
    @classmethod
    def update_colorbar(cls, plane, colorbar):
        if plane == 'xy':
            cls.colorbar_xy = colorbar
        elif plane == 'xz':
            cls.colorbar_xz = colorbar
        elif plane == 'zy':
            cls.colorbar_zy = colorbar

    @classmethod
    def update_cax(cls, plane, cax):
        if plane == 'xy':
            cls.cax_xy = cax
        elif plane == 'xz':
            cls.cax_xz = cax
        elif plane == 'zy':
            cls.cax_zy = cax
            
    @classmethod
    def update_fig(cls, plane, fig):
        if plane == 'xy':
            cls.fig_xy = fig
        elif plane == 'xz':
            cls.fig_xz = fig
        elif plane == 'zy':
            cls.fig_zy = fig

    @classmethod
    def update_ax_coord(cls, plane, ax):
        if plane == 'xy':
            cls.ax_coord_xy = ax
        elif plane == 'xz':
            cls.ax_coord_xz = ax
        elif plane == 'zy':
            cls.ax_coord_zy = ax
            
    @classmethod
    def update_xrange_input(cls, plane, xmin_input, xmax_input):
        if plane == 'xy':
            cls.xmin_input_xy = xmin_input
            cls.xmax_input_xy = xmax_input
        elif plane == 'xz':
            cls.xmin_input_xz = xmin_input
            cls.xmax_input_xz = xmax_input
        elif plane == 'zy':
            cls.xmin_input_zy = xmin_input
            cls.xmax_input_zy = xmax_input

    @classmethod
    def update_yrange_input(cls, plane, ymin_input, ymax_input):
        if plane == 'xy':
            cls.ymin_input_xy = ymin_input
            cls.ymax_input_xy = ymax_input
        elif plane == 'xz':
            cls.ymin_input_xz = ymin_input
            cls.ymax_input_xz = ymax_input
        elif plane == 'zy':
            cls.ymin_input_zy = ymin_input
            cls.ymax_input_zy = ymax_input
            
    @classmethod
    def update_zrange_input(cls, plane, zmin_input, zmax_input):
        if plane == 'xy':
            cls.zmin_input_xy = zmin_input
            cls.zmax_input_xy = zmax_input
        elif plane == 'xz':
            cls.zmin_input_xz = zmin_input
            cls.zmax_input_xz = zmax_input
        elif plane == 'zy':
            cls.zmin_input_zy = zmin_input
            cls.zmax_input_zy = zmax_input
    
    @classmethod
    def update_integ_cax(cls, cax):
        cls.integ_cax.append(cax)

    @classmethod
    def remove_integ_cax(cls, cax):
        if cax in cls.integ_cax:
            index = cls.integ_cax.index(cax)
            cls.integ_cax.pop(index)

            
    @classmethod
    def update_integ_colorbar(cls, colorbar):
        cls.integ_colorbar.append(colorbar)

    @classmethod
    def remove_integ_colorbar(cls, colorbar):
        if colorbar in cls.integ_colorbar:
            index = cls.integ_colorbar.index(colorbar)
            cls.integ_colorbar.pop(index)

    @classmethod
    def update_ch_colorbar(cls, colorbar):
        cls.ch_colorbar = colorbar
        
    @classmethod
    def update_ch_cax(cls, cax):
        cls.ch_cax = cax



"""Utility helpers for working with large FITS data arrays."""
from __future__ import annotations

import math
import os
import sys
from typing import Any, Tuple

import numpy as np


def _get_total_ram_bytes() -> int | None:
    """Return total physical RAM in bytes, or None if unavailable.

    Works on macOS, Linux, and Windows without external dependencies.
    The call reads kernel metadata and completes in microseconds.
    """
    try:
        if sys.platform == "win32":
            import ctypes

            class _MEMORYSTATUSEX(ctypes.Structure):
                _fields_ = [
                    ("dwLength", ctypes.c_ulong),
                    ("dwMemoryLoad", ctypes.c_ulong),
                    ("ullTotalPhys", ctypes.c_ulonglong),
                    ("ullAvailPhys", ctypes.c_ulonglong),
                    ("ullTotalPageFile", ctypes.c_ulonglong),
                    ("ullAvailPageFile", ctypes.c_ulonglong),
                    ("ullTotalVirtual", ctypes.c_ulonglong),
                    ("ullAvailVirtual", ctypes.c_ulonglong),
                    ("ullAvailExtendedVirtual", ctypes.c_ulonglong),
                ]

            stat = _MEMORYSTATUSEX()
            stat.dwLength = ctypes.sizeof(stat)
            if ctypes.windll.kernel32.GlobalMemoryStatusEx(ctypes.byref(stat)):
                return int(stat.ullTotalPhys)
            return None
        else:
            # POSIX (macOS, Linux)
            pages = os.sysconf("SC_PHYS_PAGES")
            page_size = os.sysconf("SC_PAGE_SIZE")
            if pages > 0 and page_size > 0:
                return pages * page_size
            return None
    except (OSError, ValueError, AttributeError):
        return None


def _auto_large_data_threshold_mb() -> int:
    """Return a Large Data Mode threshold in MiB based on system RAM.

    Heuristic: 25% of total RAM, clamped to [2048, 8192] MiB.
    Falls back to 4096 MiB if RAM detection fails.
    """
    total = _get_total_ram_bytes()
    if total is None:
        return 4096
    quarter_mb = total // 4 // (1024 * 1024)
    return max(2048, min(8192, int(quarter_mb)))


# Default thresholds for data handling. Adjust here if you need different limits.
DEFAULT_MEMMAP_THRESHOLD_MB = 1024
MEMMAP_THRESHOLD_BYTES = DEFAULT_MEMMAP_THRESHOLD_MB * 1024 * 1024

# Large Data Mode activates downsampled display and sampled statistics.
# Threshold auto-scales to 25% of system RAM (clamped 2–8 GiB).
DEFAULT_LARGE_DATA_MODE_THRESHOLD_MB = _auto_large_data_threshold_mb()
LARGE_DATA_MODE_THRESHOLD_BYTES = DEFAULT_LARGE_DATA_MODE_THRESHOLD_MB * 1024 * 1024

# FITS scaling keywords disable the fast memmap path in astropy.  When lazy
# scaling is not available (legacy path), use a lower threshold.
DEFAULT_LARGE_DATA_NO_MEMMAP_THRESHOLD_MB = 2048
LARGE_DATA_MODE_NO_MEMMAP_THRESHOLD_BYTES = (
    DEFAULT_LARGE_DATA_NO_MEMMAP_THRESHOLD_MB * 1024 * 1024
)

# Size threshold (bytes) above which load_fits keeps a raw memmap and applies
# BZERO/BSCALE lazily per-slice instead of letting astropy materialise the
# entire scaled array.  Files below this use the traditional eager path
# (memmap=False with astropy handling BZERO/BSCALE/BLANK), which is fast
# enough for files that fit comfortably in memory.
LAZY_SCALING_THRESHOLD_BYTES = 2 * 1024 * 1024 * 1024  # 2 GiB

# Keep browse-mode display work near screen resolution.
DEFAULT_LARGE_DATA_DISPLAY_MAX_DIM = 2048

# Default sampling target when deriving quick statistics from large arrays.
# Lowering the cap keeps lazy-loading responsive on very large cubes.
_DEFAULT_MAX_SAMPLES = 1_000_000


class LazyScaledArray:
    """Wraps a raw FITS memmap and applies BZERO/BSCALE lazily per-slice.

    This avoids the astropy behaviour of materialising the entire scaled float
    array when BZERO/BSCALE/BLANK keywords are present, keeping the file
    memory-mapped.  Scaling is applied only when a 2-D or smaller slice is
    extracted (the common display/analysis access pattern).

    Higher-dimensional slices (ndim >= 3) are returned as another
    ``LazyScaledArray`` so that cube → sub-cube access chains remain lazy.
    """

    def __init__(self, raw, bzero: float = 0.0, bscale: float = 1.0, blank=None):
        self._raw = raw
        self._bzero = float(bzero)
        self._bscale = float(bscale)
        self._blank = blank  # integer BLANK value or None

    # ------------------------------------------------------------------
    # numpy-compatible attribute interface
    # ------------------------------------------------------------------
    @property
    def shape(self):
        return self._raw.shape

    @property
    def ndim(self):
        return self._raw.ndim

    @property
    def dtype(self):
        return np.dtype(np.float64)

    @property
    def size(self):
        return self._raw.size

    @property
    def nbytes(self):
        return self._raw.size * 8  # float64

    @property
    def flags(self):
        return self._raw.flags

    def __len__(self):
        return len(self._raw)

    def __repr__(self):
        return (
            f"LazyScaledArray(shape={self.shape}, bzero={self._bzero}, "
            f"bscale={self._bscale}, blank={self._blank})"
        )

    def __copy__(self):
        """Shallow copy — share the underlying memmap."""
        return LazyScaledArray(self._raw, self._bzero, self._bscale, self._blank)

    def __deepcopy__(self, memo):
        """Deep copy without duplicating the memory-mapped buffer.

        The raw memmap is a read-only view into the file and can be safely
        shared across copies.  Duplicating it would defeat the purpose of
        lazy scaling and cause multi-gigabyte allocations.
        """
        return LazyScaledArray(self._raw, self._bzero, self._bscale, self._blank)

    # ------------------------------------------------------------------
    # Scaling helpers
    # ------------------------------------------------------------------
    def _apply_scaling(self, raw_data):
        """Apply BZERO/BSCALE to raw data, returning a float64 array."""
        result = np.array(raw_data, dtype=np.float64)
        if self._bscale != 1.0:
            result *= self._bscale
        if self._bzero != 0.0:
            result += self._bzero
        if self._blank is not None:
            with np.errstate(invalid="ignore"):
                blank_mask = raw_data == self._blank
                if np.any(blank_mask):
                    result[blank_mask] = np.nan
        return result

    # ------------------------------------------------------------------
    # Data access
    # ------------------------------------------------------------------
    def __getitem__(self, key):
        raw = self._raw[key]
        if isinstance(raw, np.ndarray):
            if raw.ndim >= 3:
                return LazyScaledArray(raw, self._bzero, self._bscale, self._blank)
            return self._apply_scaling(raw)
        # Scalar access
        if self._blank is not None and raw == self._blank:
            return np.nan
        return float(raw) * self._bscale + self._bzero

    def __array__(self, dtype=None):
        """Support ``np.asarray(obj)`` — materialises the entire array."""
        result = self._apply_scaling(self._raw)
        if dtype is not None:
            return result.astype(dtype)
        return result

    def reshape(self, *args, **kwargs):
        """Reshape the underlying raw array (view when possible)."""
        return LazyScaledArray(
            self._raw.reshape(*args, **kwargs),
            self._bzero,
            self._bscale,
            self._blank,
        )

    # ------------------------------------------------------------------
    # View-like operations on the raw memmap
    # ------------------------------------------------------------------
    def _raw_view_op(self, func, *args, **kwargs):
        """Apply a numpy view operation (flip, expand_dims, …) to the raw array."""
        return LazyScaledArray(
            func(self._raw, *args, **kwargs),
            self._bzero,
            self._bscale,
            self._blank,
        )

    def astype(self, dtype, **kwargs):
        """Return a materialised array cast to *dtype*."""
        return self.__array__().astype(dtype, **kwargs)


def is_lazy_scaled(obj) -> bool:
    """Return True if *obj* is a :class:`LazyScaledArray`."""
    return isinstance(obj, LazyScaledArray)


def estimate_array_nbytes(array: np.ndarray | np.memmap | LazyScaledArray | None) -> int | None:
    """Return the approximate number of bytes for the given array.

    The helper works for ``numpy.ndarray``, ``numpy.memmap``, and
    ``LazyScaledArray`` instances.  ``None`` is returned if the shape or
    dtype is not available.
    """
    if array is None:
        return None

    if isinstance(array, LazyScaledArray):
        return int(array._raw.size) * array._raw.dtype.itemsize

    try:
        return int(array.size) * array.dtype.itemsize
    except (AttributeError, TypeError, ValueError):
        return None


def format_nbytes(num_bytes: int | None) -> str:
    """Return a compact human-readable byte string."""
    if num_bytes is None:
        return "unknown size"
    if num_bytes < 0:
        num_bytes = 0

    value = float(num_bytes)
    units = ("B", "KiB", "MiB", "GiB", "TiB", "PiB")
    unit_index = 0
    while value >= 1024.0 and unit_index < len(units) - 1:
        value /= 1024.0
        unit_index += 1
    if unit_index == 0:
        return f"{int(value)} {units[unit_index]}"
    return f"{value:.1f} {units[unit_index]}"


def _parse_header_float(header: Any, key: str, default: float) -> float | None:
    """Return a numeric header value, or ``None`` when parsing fails."""
    try:
        if key not in header:
            return float(default)
        value = header[key]
    except Exception:
        return None

    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def header_has_scaling_keywords(header: Any | None) -> bool:
    """Return True when FITS data requires scaling or BLANK handling.

    BLANK is only meaningful for integer BITPIX (>0).  For floating-point
    data (BITPIX < 0) the FITS standard says BLANK must be ignored, and
    astropy does exactly that, so it never prevents memory-mapping.
    """
    if header is None:
        return False

    try:
        bitpix = int(header.get("BITPIX", 0))
    except (TypeError, ValueError):
        bitpix = 0

    try:
        if "BLANK" in header and bitpix > 0:
            return True
    except Exception:
        return True

    bzero = _parse_header_float(header, "BZERO", 0.0)
    bscale = _parse_header_float(header, "BSCALE", 1.0)
    if bzero is None or bscale is None:
        return True

    return not (bzero == 0.0 and bscale == 1.0)


def _threshold_bytes_from_config(
    config: dict[str, Any] | None,
    key: str,
    default_mb: int,
) -> int:
    """Return a threshold in bytes, with optional config override in MiB."""
    if not isinstance(config, dict):
        return int(default_mb) * 1024 * 1024

    value = config.get(key, default_mb)
    try:
        value_mb = int(value)
    except (TypeError, ValueError):
        value_mb = int(default_mb)

    if value_mb <= 0:
        value_mb = int(default_mb)
    return value_mb * 1024 * 1024


def build_large_data_profile(
    array: np.ndarray | np.memmap | None,
    *,
    header: Any | None = None,
    config: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Describe whether the array should surface as Large Data Mode."""
    approx_bytes = estimate_array_nbytes(array)
    has_scaling_keywords = header_has_scaling_keywords(header)
    regular_threshold_bytes = _threshold_bytes_from_config(
        config,
        "large_data_mode_threshold_mb",
        DEFAULT_LARGE_DATA_MODE_THRESHOLD_MB,
    )
    no_memmap_threshold_bytes = _threshold_bytes_from_config(
        config,
        "large_data_no_memmap_threshold_mb",
        DEFAULT_LARGE_DATA_NO_MEMMAP_THRESHOLD_MB,
    )
    threshold_bytes = (
        no_memmap_threshold_bytes
        if has_scaling_keywords
        else regular_threshold_bytes
    )
    enabled = approx_bytes is not None and approx_bytes >= threshold_bytes

    if approx_bytes is None:
        reason = "Large Data Mode unavailable because data size could not be estimated."
    elif has_scaling_keywords and enabled:
        reason = (
            f"estimated size {format_nbytes(approx_bytes)} exceeds "
            f"{format_nbytes(threshold_bytes)} and FITS scaling keywords limit "
            "the fast memmap path"
        )
    elif enabled:
        reason = (
            f"estimated size {format_nbytes(approx_bytes)} exceeds "
            f"{format_nbytes(threshold_bytes)}"
        )
    else:
        reason = (
            f"estimated size {format_nbytes(approx_bytes)} is within the "
            f"{format_nbytes(threshold_bytes)} Large Data Mode threshold"
        )

    return {
        "enabled": bool(enabled),
        "estimated_size_bytes": approx_bytes,
        "estimated_size_text": format_nbytes(approx_bytes),
        "threshold_bytes": threshold_bytes,
        "threshold_text": format_nbytes(threshold_bytes),
        "has_scaling_keywords": bool(has_scaling_keywords),
        "reason": reason,
    }


def sanitize_slice(array: np.ndarray) -> np.ndarray:
    """Replace invalid values (< -100000) and infinities with NaN in-place.

    Intended for per-slice use so that the full cube is never scanned at load
    time.  Returns *array* (possibly converted to float) for convenience.
    """
    if array is None or array.size == 0:
        return array
    with np.errstate(invalid="ignore"):
        bad = (array < -100000) | ~np.isfinite(array)
    if np.any(bad):
        if not np.issubdtype(array.dtype, np.floating):
            array = array.astype(np.float32)
        array[bad] = np.nan
    return array


def downsample_2d_for_display(
    array: np.ndarray | np.memmap | LazyScaledArray,
    *,
    max_dimension: int = DEFAULT_LARGE_DATA_DISPLAY_MAX_DIM,
) -> np.ndarray:
    """Return a 2-D display array capped to roughly ``max_dimension`` pixels per axis."""
    max_dimension = max(1, int(max_dimension))

    # For LazyScaledArray, stride the raw memmap first, then scale only the
    # small result so the full array is never materialised.
    if isinstance(array, LazyScaledArray) and array.ndim == 2:
        height, width = array.shape
        step_y = max(1, int(math.ceil(height / max_dimension)))
        step_x = max(1, int(math.ceil(width / max_dimension)))
        if step_y == 1 and step_x == 1:
            return array._apply_scaling(array._raw)
        raw_ds = array._raw[::step_y, ::step_x]
        return array._apply_scaling(raw_ds)

    arr = np.asanyarray(array)
    if arr.ndim != 2:
        return arr

    height, width = arr.shape
    step_y = max(1, int(math.ceil(height / max_dimension)))
    step_x = max(1, int(math.ceil(width / max_dimension)))
    if step_y == 1 and step_x == 1:
        if getattr(arr, "flags", None) is not None and arr.flags.c_contiguous:
            return arr
        return np.array(arr, copy=True)

    return np.array(arr[::step_y, ::step_x], copy=True)


def _select_sample(flat: np.ndarray, max_samples: int) -> np.ndarray:
    """Select a representative 1-D sample from ``flat`` with at most ``max_samples`` values."""
    total = flat.size
    if total == 0:
        return flat

    if total <= max_samples:
        return flat

    step = max(1, total // max_samples)
    sample = flat[::step]
    if sample.size > max_samples:
        sample = sample[:max_samples]

    # Ensure the last element is included so we do not miss a possible extremum.
    if sample.size == 0 or sample[-1] != flat[-1]:
        sample = np.concatenate((sample, flat[-1:]))

    return sample


def fast_nanminmax(
    array: np.ndarray | np.memmap | LazyScaledArray,
    max_samples: int = _DEFAULT_MAX_SAMPLES,
) -> Tuple[float, float]:
    """Return an approximate ``(nanmin, nanmax)`` pair for ``array``.

    The function is designed for very large arrays backed by memory maps. Instead
    of materialising the full dataset, a sub-sample is inspected. The result is
    sufficient for display initialisation while avoiding multi-gigabyte scans.

    For :class:`LazyScaledArray` instances the sample is drawn from the raw
    memmap and then scaled, so the full array is never materialised.
    """
    if isinstance(array, LazyScaledArray):
        raw_flat = array._raw.reshape(-1)
        raw_sample = _select_sample(raw_flat, max_samples)
        sample = array._apply_scaling(raw_sample)
    else:
        arr = np.asanyarray(array)
        if arr.size == 0:
            return (math.nan, math.nan)
        flat = arr.reshape(-1)
        sample = _select_sample(flat, max_samples)

    with np.errstate(all="ignore"):
        finite = sample[np.isfinite(sample)]
        if finite.size == 0:
            return (math.nan, math.nan)

        return (float(np.min(finite)), float(np.max(finite)))



from matplotlib.backends.backend_qtagg import NavigationToolbar2QT
import logging
import os
import yaml

from takefits.app_paths import app_config_path
from PySide6 import QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtGui import QFontMetrics, QIcon, QCursor
from PySide6.QtWidgets import QFileDialog, QMessageBox

from takefits.tools.color_scale import ColorMode

logger = logging.getLogger(__name__)

class MyNavigationToolbar(NavigationToolbar2QT):
    _nav_mode_sync_in_progress = False

    def __init__(self, canvas, parent, plane,ax, color_mode = None, default_image_name="figure", default_image_ext="pdf"):
        # NavigationToolbar2QT.__init__ may call set_history_buttons(), so fields
        # referenced there must exist before super().__init__().
        self.canvas = canvas
        self.parent = parent
        self.plane = plane
        self.ax = ax
        self.color_mode = color_mode
        self.default_image_name = default_image_name
        self.default_image_ext = default_image_ext
        self._cursor_override_active = False
        self._external_can_back = False
        self._external_can_forward = False
        super().__init__(canvas, parent)
        loc_label = getattr(self, "locLabel", None)
        if loc_label is not None:
            # Keep multiline coordinate text readable on narrow windows.
            loc_label.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignTop)
            loc_label.setTextFormat(Qt.TextFormat.PlainText)
            loc_label.setWordWrap(False)
            two_line_height = (loc_label.fontMetrics().lineSpacing() * 2) + 4
            loc_label.setMinimumHeight(two_line_height)
            loc_label.setMaximumHeight(two_line_height)
            loc_label.setSizePolicy(
                QtWidgets.QSizePolicy.Policy.Expanding,
                QtWidgets.QSizePolicy.Policy.Preferred,
            )
        self._apply_cursor_for_mode()


    def configure_subplots(self, *args, **kwargs):
        """
        If a dialog is already open, just bring it to front and update window size.
        Otherwise, create a new one.
        """
        # 1) Check if a dialog is already open and visible
        if self._subplot_dialog is not None and self._subplot_dialog.isVisible():
            self._subplot_dialog.width_spin.setValue(int(self.parent.window().width()))
            self._subplot_dialog.height_spin.setValue(int(self.parent.window().height()))
            self._subplot_dialog.raise_()
            self._subplot_dialog.activateWindow()
            return self._subplot_dialog
        else:
            # 2) Otherwise, close the old one if it exists (but hidden), then recreate
            if self._subplot_dialog is not None:
                self._subplot_dialog.close()
                self._subplot_dialog = None
    
            self._subplot_dialog = MySubplotToolQt(self.canvas.figure, self.parent)
            self.canvas.mpl_connect("close_event", lambda e: self._subplot_dialog.reject())
    
            # When the parent window is destroyed, also close the dialog
            self.parent.window().destroyed.connect(self._subplot_dialog.close)
    
            # Update spinboxes with current window size
            self._subplot_dialog.update_from_current_subplotpars()
    
            # Show new dialog
            self._subplot_dialog.show()
            return self._subplot_dialog
            

    def zoom(self, *args):
        super().zoom(*args)
        self._apply_cursor_for_mode()
        self._sync_navigation_mode('zoom')

    def set_message(self, s):
        text = str(s or "")
        text = text.replace("\r\n", "\n").replace("\r", "\n")
        normalized = []
        for raw in text.split("\n"):
            line = " ".join(str(raw).split()).strip()
            if line:
                normalized.append(line)
        if not normalized:
            super().set_message("")
            return
        if len(normalized) > 2:
            normalized = normalized[:2]
        super().set_message("\n".join(normalized))

    def pan(self, *args):
        super().pan(*args)
        self._apply_cursor_for_mode()
        self._sync_navigation_mode('pan')

    def _iter_linked_toolbars(self):
        target = self._resolve_root_viewer()
        if target is None:
            return []
        viewers = [target] + [v for v in list(getattr(target, 'subwindows', []) or []) if v is not None]
        toolbars = []
        for viewer in viewers:
            toolbar = getattr(viewer, 'toolbar', None)
            if isinstance(toolbar, MyNavigationToolbar) and getattr(toolbar, 'color_mode', None) is None:
                toolbars.append(toolbar)
        return toolbars

    def _sync_navigation_mode(self, source_mode):
        if self.color_mode is not None:
            return
        if MyNavigationToolbar._nav_mode_sync_in_progress:
            return
        desired_mode = 'pan/zoom' if source_mode == 'pan' else 'zoom rect'
        mode_now = str(getattr(self, 'mode', '') or '').lower()
        should_enable = (mode_now == desired_mode)

        linked_toolbars = [tb for tb in self._iter_linked_toolbars() if tb is not self]
        if not linked_toolbars:
            return

        MyNavigationToolbar._nav_mode_sync_in_progress = True
        try:
            for toolbar in linked_toolbars:
                current_mode = str(getattr(toolbar, 'mode', '') or '').lower()
                if source_mode == 'pan':
                    if should_enable and current_mode != 'pan/zoom':
                        toolbar.pan()
                    elif (not should_enable) and current_mode == 'pan/zoom':
                        toolbar.pan()
                elif source_mode == 'zoom':
                    if should_enable and current_mode != 'zoom rect':
                        toolbar.zoom()
                    elif (not should_enable) and current_mode == 'zoom rect':
                        toolbar.zoom()
        finally:
            MyNavigationToolbar._nav_mode_sync_in_progress = False

    def sync_navigation_mode_from_linked(self):
        """Adopt current pan/zoom mode from already-open linked viewers."""
        if self.color_mode is not None:
            return
        if MyNavigationToolbar._nav_mode_sync_in_progress:
            return

        linked_toolbars = [tb for tb in self._iter_linked_toolbars() if tb is not self]
        desired = None
        for toolbar in linked_toolbars:
            mode = str(getattr(toolbar, 'mode', '') or '').lower()
            if mode == 'pan/zoom':
                desired = 'pan'
                break
            if mode == 'zoom rect' and desired is None:
                desired = 'zoom'

        current_mode = str(getattr(self, 'mode', '') or '').lower()
        if desired == 'pan' and current_mode != 'pan/zoom':
            self.pan()
        elif desired == 'zoom' and current_mode != 'zoom rect':
            self.zoom()


    def release_zoom(self, event):
        super().release_zoom(event)
        self.get_current_lim(event)
        self._record_shared_view_history("zoom")
        self._apply_cursor_for_mode()
        self._refresh_overlay_after_nav()
        self._notify_view_history_changed()

        #self.parent._background = self.canvas.copy_from_bbox(self.parent.overlay_ax.bbox)


    def release_pan(self, event):
        super().release_pan(event)
        self.get_current_lim(event)
        self._record_shared_view_history("pan")
        self._apply_cursor_for_mode()
        self._refresh_overlay_after_nav()
        self._notify_view_history_changed()

    def back(self, *args):
        if self.color_mode is None:
            target = self._resolve_root_viewer()
            callback = getattr(target, "view_back", None)
            if callable(callback):
                try:
                    callback()
                finally:
                    self._apply_cursor_for_mode()
                return
        else:
            callback = getattr(self.parent, "view_back", None)
            if callable(callback):
                try:
                    callback()
                finally:
                    self._apply_cursor_for_mode()
                return
        super().back(*args)
        self.get_current_lim(None)
        self._apply_cursor_for_mode()
        self._refresh_overlay_after_nav()
        self._notify_view_history_changed()

    def forward(self, *args):
        if self.color_mode is None:
            target = self._resolve_root_viewer()
            callback = getattr(target, "view_forward", None)
            if callable(callback):
                try:
                    callback()
                finally:
                    self._apply_cursor_for_mode()
                return
        else:
            callback = getattr(self.parent, "view_forward", None)
            if callable(callback):
                try:
                    callback()
                finally:
                    self._apply_cursor_for_mode()
                return
        super().forward(*args)
        self.get_current_lim(None)
        self._apply_cursor_for_mode()
        self._refresh_overlay_after_nav()
        self._notify_view_history_changed()

    def set_history_buttons(self):
        super().set_history_buttons()
        if getattr(self, "color_mode", None) is None:
            target = self._resolve_root_viewer()
            state_getter = getattr(target, "_shared_view_history_state", None)
            if callable(state_getter):
                try:
                    can_back, can_forward = state_getter()
                    self.set_external_history_state(can_back, can_forward)
                except Exception:
                    self._apply_external_history_state()
        else:
            state_getter = getattr(self.parent, "_view_history_state", None)
            if callable(state_getter):
                try:
                    can_back, can_forward = state_getter()
                    self.set_external_history_state(can_back, can_forward)
                except Exception:
                    self._apply_external_history_state()
        self._notify_view_history_changed()

    def press_pan(self, event):
        super().press_pan(event)
        self._apply_cursor_for_mode(during_pan=True)
        # optional? no draw on move

    def drag_pan(self, motion_event):
        super().drag_pan(motion_event)
        self._apply_cursor_for_mode(during_pan=True)

    def get_current_lim(self, event=None):
        if event is not None and event.inaxes and event.inaxes.get_gid() == "colorbar":
            return
        if self.ax is None:
            return

        xlim = self.ax.get_xlim()
        ylim = self.ax.get_ylim()
        self._apply_world_ranges_from_limits(xlim, ylim)

    def _resolve_root_viewer(self):
        viewer = getattr(self, 'parent', None)
        if viewer is None:
            return None
        main_viewer = getattr(viewer, 'main_viewer', None)
        if main_viewer is not None:
            return main_viewer
        return viewer

    def _apply_world_ranges_from_limits(self, xlim, ylim):
        if self.color_mode == None:
            target = self._resolve_root_viewer()
            if target is None:
                return
            begin_batch = getattr(target, "_begin_view_history_batch", None)
            end_batch = getattr(target, "_end_view_history_batch", None)
            if callable(begin_batch):
                begin_batch()
            range_panel = getattr(target, 'range_panel', None)
            try:
                if range_panel is not None and hasattr(range_panel, 'update_ranges'):
                    range_panel.update_ranges(self.plane, xlim, ylim)
                if hasattr(target, 'update_ranges'):
                    target.update_ranges(self.plane, xlim, ylim)
                if self.plane == 'xy':
                    if hasattr(target, 'set_x_range'):
                        target.set_x_range()
                    if hasattr(target, 'set_y_range'):
                        target.set_y_range()
                elif self.plane == 'xz':
                    if hasattr(target, 'set_x_range'):
                        target.set_x_range()
                    if hasattr(target, 'set_z_range'):
                        target.set_z_range()
                elif self.plane == 'zy':
                    if hasattr(target, 'set_y_range'):
                        target.set_y_range()
                    if hasattr(target, 'set_z_range'):
                        target.set_z_range()
            finally:
                if callable(end_batch):
                    end_batch()
        elif self.color_mode == ColorMode.INTEG or self.color_mode == ColorMode.CHANNEL:
            self.parent.update_ranges(self.plane, xlim, ylim)

    def _notify_view_history_changed(self):
        if self.color_mode is None:
            target = self._resolve_root_viewer()
        else:
            target = getattr(self, "parent", None)
        if target is None:
            return
        refresh_fn = getattr(target, "_refresh_view_navigation_actions", None)
        if callable(refresh_fn):
            try:
                refresh_fn()
            except Exception:
                pass

    def home(self, *args):
        if self.color_mode == None:
            target_viewer = self._resolve_root_viewer()

            if target_viewer is not None and hasattr(target_viewer, 'reset_all_ranges'):
                target_viewer.reset_all_ranges()
        elif self.color_mode == ColorMode.INTEG or self.color_mode == ColorMode.CHANNEL:
            self.parent.set_full_range()
        self._apply_cursor_for_mode()
        self._notify_view_history_changed()

    def set_external_history_state(self, can_back: bool, can_forward: bool):
        self._external_can_back = bool(can_back)
        self._external_can_forward = bool(can_forward)
        self._apply_external_history_state()

    def _apply_external_history_state(self):
        actions = getattr(self, "_actions", None)
        if not isinstance(actions, dict):
            return
        back_action = actions.get("back")
        forward_action = actions.get("forward")
        if back_action is not None:
            try:
                back_action.setEnabled(bool(self._external_can_back))
            except Exception:
                pass
        if forward_action is not None:
            try:
                forward_action.setEnabled(bool(self._external_can_forward))
            except Exception:
                pass

    def _record_shared_view_history(self, reason: str):
        if self.color_mode is not None:
            return
        target = self._resolve_root_viewer()
        recorder = getattr(target, "_record_shared_view_history", None)
        if callable(recorder):
            try:
                recorder(reason=reason)
            except Exception:
                pass

    def _apply_cursor_for_mode(self, *, during_pan=False):
        app = QtWidgets.QApplication.instance()
        if app is None:
            return

        if self._cursor_override_active:
            self._clear_cursor_override(app)

        cursor = None
        if self.mode == 'zoom rect':
            cursor = QCursor(Qt.CursorShape.CrossCursor)
        elif self.mode == 'pan/zoom':
            cursor = QCursor(Qt.CursorShape.SizeAllCursor)

        if cursor is not None:
            app.setOverrideCursor(cursor)
            self._cursor_override_active = True

    def _refresh_overlay_after_nav(self):
        parent = getattr(self, 'parent', None)
        plane = getattr(parent, 'plane', None) or getattr(self, 'plane', None)

        refreshed = False
        if parent is not None:
            refresh_fn = getattr(parent, 'redraw_overlay_for_plane', None)
            if callable(refresh_fn) and plane is not None:
                try:
                    refresh_fn(plane)
                    refreshed = True
                except Exception:
                    refreshed = False

        marker_manager = getattr(parent, 'marker_manager', None)
        if marker_manager is None:
            marker_manager = getattr(self, 'marker_manager', None)

        if not refreshed and marker_manager is not None and plane is not None:
            try:
                marker_manager.draw_markers_for_blit(plane)
                marker_manager.redraw_plane(plane)
            except Exception:
                pass

    def _clear_cursor_override(self, app=None):
        if not self._cursor_override_active:
            return
        if app is None:
            app = QtWidgets.QApplication.instance()
        if app is None:
            self._cursor_override_active = False
            return
        try:
            app.restoreOverrideCursor()
        finally:
            self._cursor_override_active = False

    def __del__(self):
        self._clear_cursor_override()

    def onclick(self, event):
        if self.mode == 'zoom rect': pass
        
    def save_figure(self):
        current_dir = os.getcwd()
        desktop_dir = os.path.join(os.path.expanduser("~"), "Desktop")
        initial_dir = desktop_dir if os.access(desktop_dir, os.W_OK) else current_dir
        if not self.default_image_name:
            if hasattr(self.parent, "filename") and self.parent.filename:
                self.default_image_name = self.parent.filename
            else:
                self.default_image_name = "figure"
        if self.default_image_name.endswith(".fits"):
            self.default_image_name = self.default_image_name[:-5]

        try:
            self.default_image_ext = f"{self.parent.integ_mode}.{self.default_image_ext}"
        except: pass
        default_filename = f"{self.default_image_name}.{self.default_image_ext}"

        path, _ = QFileDialog.getSaveFileName(
            self.canvas.parent(),
            "Save Figure",
            os.path.join(initial_dir, default_filename),
            "PDF Files (*.pdf);;EPS Files (*.eps);;SVG Files (*.svg);;PNG Files (*.png);;JPEG Files (*.jpg *.jpeg);;All Files (*)"
        )
        if path:
            visibility = []
            animated = []
            try:
                for patch in self.ax.patches:
                    visibility.append(patch.get_visible())
                    animated.append(getattr(patch, 'get_animated', lambda: False)())
                    patch.set_visible(True)
                    if hasattr(patch, 'set_animated'):
                        patch.set_animated(False)
                self.canvas.figure.savefig(path, transparent=True, dpi=300)
                filename = os.path.basename(path)
                self.show_save_success_message(path, filename)
            finally:
                for patch, vis, anim in zip(self.ax.patches, visibility, animated):
                    patch.set_visible(vis)
                    if hasattr(patch, 'set_animated'):
                        patch.set_animated(anim)
                self._refresh_overlays()
            # Restore a reasonable status message after saving, if available.
            # Use the Axes' format_coord (expects data coords), not the canvas.
            try:
                if self.ax is not None and hasattr(self.ax, "format_coord"):
                    xmid = sum(self.ax.get_xlim()) / 2.0
                    ymid = sum(self.ax.get_ylim()) / 2.0
                    message = self.ax.format_coord(xmid, ymid)
                    if message:
                        self.set_message(message)
            except Exception:
                # Silently ignore any issues restoring the message.
                pass

    def show_save_success_message(self, path, filename):
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Icon.Information)
        msg.setText(f'"{filename}" was saved successfully at:\n{path}')
        #print(f'\n\n"{filename}" was saved successfully at:\n{path}')
        msg.setStandardButtons(QMessageBox.StandardButton.Ok)
        msg.exec()

    def set_axes(self, new_ax):
        self.ax = new_ax

    def _refresh_overlays(self):
        refresh = getattr(self.parent, 'redraw_main_overlay_and_blit', None)
        if callable(refresh):
            refresh()



class MySubplotToolQt(QtWidgets.QDialog):
    def __init__(self, targetfig, parent):
        super().__init__(parent)
        self.subplot_params_file = app_config_path('subplot_params.yaml')
        self._parent = parent
        self.setWindowIcon(QIcon("matplotlib.png"))
        self.setObjectName("SubplotTool")
        self._spinboxes = {}
        main_layout = QtWidgets.QHBoxLayout()
        main_layout.setContentsMargins(10, 0, 10, 10)
        self.setLayout(main_layout)
        for group, spinboxes, buttons in [
                ("Borders",
                 ["top", "bottom", "left", "right"],
                 [("Export values", self._export_values)]),
                ("Spacings",
                 ["hspace", "wspace"],
                 [("Tight layout", self._tight_layout),
                  ("Reset", self._reset),
                  ("Close", self.close)])]:
            layout = QtWidgets.QVBoxLayout()
            main_layout.addLayout(layout)
            box = QtWidgets.QGroupBox(group)
            layout.addWidget(box)
            inner = QtWidgets.QFormLayout(box)
            for name in spinboxes:
                self._spinboxes[name] = spinbox = QtWidgets.QDoubleSpinBox()
                spinbox.setRange(0, 1)
                spinbox.setDecimals(3)
                spinbox.setSingleStep(0.005)
                spinbox.setKeyboardTracking(False)
                spinbox.valueChanged.connect(self._on_value_changed)
                inner.addRow(name, spinbox)
            layout.addStretch(1)
            for name, method in buttons:
                button = QtWidgets.QPushButton(name)
                # Don't trigger on <enter>, which is used to input values.
                button.setAutoDefault(False)
                button.clicked.connect(method)
                layout.addWidget(button)
                #if name == "Close":
                #    button.setFocus()
        self._figure = targetfig
        self._defaults = {}
        self._export_values_dialog = None
        
        
        window_container = QtWidgets.QWidget()
        window_container.setObjectName("CustomWindowSizeContainer")
        window_container_layout = QtWidgets.QVBoxLayout(window_container)
        window_container_layout.setContentsMargins(0, 0, 0, 0)
        window_container_layout.setSpacing(5)

        window_group_box = QtWidgets.QGroupBox("Window")
        window_group_box.setFixedHeight(100)
        window_layout = QtWidgets.QGridLayout()
        window_layout.setContentsMargins(10, 0, 10, 0)
        window_layout.setSpacing(5)
        window_group_box.setLayout(window_layout)

        width_label = QtWidgets.QLabel("width")
        self.width_spin = QtWidgets.QSpinBox()
        self.width_spin.setRange(499, 9999)
        self.width_spin.setValue(int(parent.width()))
        self.width_spin.setObjectName("widthSpin")

        height_label = QtWidgets.QLabel("height")
        self.height_spin = QtWidgets.QSpinBox()
        self.height_spin.setRange(115, 9999)
        self.height_spin.setValue(int(parent.height()))
        self.height_spin.setObjectName("heightSpin")

        window_layout.addWidget(height_label, 0, 0, alignment=Qt.AlignmentFlag.AlignLeft)
        window_layout.addWidget(self.height_spin, 0, 1, alignment=Qt.AlignmentFlag.AlignLeft)
        window_layout.addWidget(width_label, 1, 0, alignment=Qt.AlignmentFlag.AlignLeft)
        window_layout.addWidget(self.width_spin, 1, 1, alignment=Qt.AlignmentFlag.AlignLeft)

        def update_window_width():
            new_width = self.width_spin.value()
            parent.resize(new_width, int(parent.window().height()))

        def update_window_height():
            new_height = self.height_spin.value()
            parent.resize(int(parent.window().width()), new_height)

        self.width_spin.valueChanged.connect(update_window_width)
        self.height_spin.valueChanged.connect(update_window_height)

        window_container_layout.addWidget(window_group_box, alignment=Qt.AlignmentFlag.AlignBottom)

        import_button = QtWidgets.QPushButton("Import values")
        import_button.setAutoDefault(False)
        import_button.clicked.connect(lambda _=False: self._import_values(show_message=True))
        window_container_layout.addWidget(import_button, alignment=Qt.AlignmentFlag.AlignBottom)


        dlg_layout = self.layout()
        existing_container = self.findChild(QtWidgets.QWidget, "CustomWindowSizeContainer")
        if existing_container is not None:
            dlg_layout.removeWidget(existing_container)
            existing_container.deleteLater()

        dlg_layout.insertWidget(0, window_container)

        tmp_dict = {'height': self.height_spin, 'width': self.width_spin}
        self._spinboxes = {**tmp_dict, **self._spinboxes}
        self.update_from_current_subplotpars()
        
    def update_from_current_subplotpars(self):
        spinboxes = self._spinboxes
        allowed_keys = {"left", "right", "top", "bottom", "wspace", "hspace"}
    
        self._defaults = {} 
    
        for name, spinbox in spinboxes.items():
            if name in allowed_keys:
                self._defaults[spinbox] = getattr(self._figure.subplotpars, name)
            elif name == "width":
                current_width = self._parent.window().width()
                self._defaults[spinbox] = current_width
                spinbox.blockSignals(True)
                spinbox.setValue(current_width)
                spinbox.blockSignals(False)
            elif name == "height":
                current_height = self._parent.window().height()
                self._defaults[spinbox] = current_height
                spinbox.blockSignals(True)
                spinbox.setValue(current_height)
                spinbox.blockSignals(False)

        self._reset()

    def _export_values(self):
        # Explicitly round to 3 decimals (which is also the spinbox precision)
        # to avoid numbers of the form 0.100...001.
        self._export_values_dialog = QtWidgets.QDialog()
        layout = QtWidgets.QVBoxLayout()
        self._export_values_dialog.setLayout(layout)
        text = QtWidgets.QPlainTextEdit()
        text.setReadOnly(True)
        layout.addWidget(text)
        text.setPlainText(
            ",\n".join(
                f"{attr}={int(spinbox.value())}" if attr in {"height", "width"}
                else f"{attr}={spinbox.value():.3}"
                for attr, spinbox in self._spinboxes.items()
            )
        )
        # Adjust the height of the text widget to fit the whole text, plus
        # some padding.
        size = text.maximumSize()
        size.setHeight(
            QFontMetrics(text.document().defaultFont())
            .size(0, text.toPlainText()).height() + 20)
        text.setMaximumSize(size)
        self._export_values_dialog.show()
        
        param_dict = {}
        for attr, spinbox in self._spinboxes.items():
            if attr in {"height", "width"}:
                # Save as integer
                param_dict[attr] = int(spinbox.value())
            else:
                # Save as float with 3 decimal places
                param_dict[attr] = float(f"{spinbox.value():.3f}")
        
        # Fixed filename
        filename = self.subplot_params_file
        
        with open(filename, "w", encoding="utf-8") as f:
            yaml.dump(param_dict, f, default_flow_style=False)
        
        print(f"\033[96mExported parameters to '{filename}'\033[0m")

    def _import_values(self, show_message: bool = True):
        """
        Import parameters from the same fixed YAML file ('myparams.yaml')
        and update each spinbox accordingly.
        """
        filename = self.subplot_params_file
        
        # Attempt to load the file
        try:
            with open(filename, "r", encoding="utf-8") as f:
                param_dict = yaml.safe_load(f)
        except FileNotFoundError:
            print(f"\033[93m\033[1mFile '{filename}' not found.\033[0m")
            return
        except Exception as e:
            print(f"Failed to load '{filename}': {e}")
            return
        
        # Update spinboxes with loaded values
        for attr, value in param_dict.items():
            if attr in self._spinboxes:
                # Update the spinbox itself
                self._spinboxes[attr].blockSignals(True)
                self._spinboxes[attr].setValue(value)
                self._spinboxes[attr].blockSignals(False)
    
                # If it's width or height, also resize the parent window
                if attr == "width":
                    # if you have a separate QSpinBox for width:
                    self.width_spin.setValue(int(value))
                    #self.width_spin_dummy.setValue(float(value))
                    # Then resize parent
                    self._parent.resize(int(value), self._parent.height())
    
                elif attr == "height":
                    self.height_spin.setValue(int(value))
                    #self.height_spin_dummy.setValue(float(value))
                    self._parent.resize(self._parent.width(), int(value))
        
        # Optionally trigger any re-draw or logic after updating
        self._on_value_changed()
        if show_message:
            print(f"\033[96mImported parameters from '{filename}' successfully.\033[0m")
        else:
            logger.debug("Imported parameters from '%s' successfully.", filename)



    def _on_value_changed(self):
        spinboxes = self._spinboxes
        # Set all mins and maxes, so that this can also be used in _reset().
        for lower, higher in [("bottom", "top"), ("left", "right")]:
            spinboxes[higher].setMinimum(spinboxes[lower].value() + .001)
            spinboxes[lower].setMaximum(spinboxes[higher].value() - .001)
        
        allowed_keys = {"left", "right", "top", "bottom", "wspace", "hspace"}
        adjust_params = {attr: spinbox.value() for attr, spinbox in spinboxes.items() if attr in allowed_keys}
        self._figure.subplots_adjust(**adjust_params)
        #self._figure.subplots_adjust(
        #    **{attr: spinbox.value() for attr, spinbox in spinboxes.items()})
        self._figure.canvas.draw_idle()

    def _tight_layout(self):
        spinboxes = self._spinboxes
        self._figure.tight_layout()
        allowed_keys = {"left", "right", "top", "bottom", "wspace", "hspace"}
        for attr, spinbox in spinboxes.items():
            if attr in allowed_keys:
                spinbox.blockSignals(True)
                spinbox.setValue(vars(self._figure.subplotpars)[attr])
                spinbox.blockSignals(False)
        self._figure.canvas.draw_idle()

    def _reset(self):
        for spinbox, value in self._defaults.items():
            #if (spinbox is self._spinboxes.get('width')
            #        or spinbox is self._spinboxes.get('height')):
            #    spinbox.setRange(100, 3000)
            #else:
            #    spinbox.setRange(0, 1)
    
            spinbox.blockSignals(True)
            spinbox.setValue(value)
            spinbox.blockSignals(False)
    
        # After setting spinboxes for width/height, resize the parent window too:
        w_val = self._spinboxes['width'].value()
        h_val = self._spinboxes['height'].value()
        self.width_spin.setValue(int(w_val))
        #self.width_spin_dummy.setValue(float(w_val))
        self.height_spin.setValue(int(h_val))
        #self.height_spin_dummy.setValue(float(h_val))
        self._parent.window().resize(int(w_val), int(h_val))
    
        self._on_value_changed()
        toggle_overlays = getattr(self.parent, 'set_overlay_updates_enabled', None)
        if callable(toggle_overlays):
            toggle_overlays(False)



"""
ViewerState: Per-plane state container for FITSViewer instances.

Holds all matplotlib objects, UI elements, crosshairs, and state flags
for a single viewer plane (xy, xz, or zy).
"""
from typing import Optional, Any
from weakref import ref as weakref


class ViewerState:
    """
    Encapsulates all per-plane state that was previously stored in Common.
    Each FITSViewer/SubWindow owns one ViewerState instance.
    """

    def __init__(self, plane: str):
        """
        Initialize state for a specific plane.

        Args:
            plane: The plane identifier ('xy', 'xz', or 'zy')
        """
        self.plane = plane

        # Matplotlib objects
        self.canvas: Any = None
        self.fig: Any = None
        self.ax: Any = None
        self.overlay_ax: Any = None
        self.im: Any = None
        self._background: Any = None

        # Crosshair lines
        self.hline: Any = None
        self.vline: Any = None
        self.cpoint: Any = None

        # Colorbar
        self.colorbar: Any = None
        self.cax: Any = None

        # Labels
        self.plabel: Any = None  # Position label (QLabel)
        self.chlabel: Any = None  # Channel label (matplotlib text)
        self.hpbw: Any = None  # Half-power beam width ellipse
        self.cursor_x: Optional[float] = None
        self.cursor_y: Optional[float] = None
        self.image_background: Any = None

        # UI elements
        self.slider: Any = None
        self.chval_box: Any = None
        self.xmin_input: Any = None
        self.xmax_input: Any = None
        self.ymin_input: Any = None
        self.ymax_input: Any = None
        self.zmin_input: Any = None
        self.zmax_input: Any = None

        # Axis metadata
        self.ax_xunit: Optional[str] = None
        self.ax_yunit: Optional[str] = None
        self.ax_xtype: Optional[str] = None
        self.ax_ytype: Optional[str] = None
        self.ax_coord: Any = None  # List of coordinate axes

        # State flags
        self.clicked: bool = False

        # Weak reference to the owning viewer (set by coordinator)
        self._viewer_ref: Optional[weakref] = None

    @property
    def viewer(self):
        """Return the owning viewer if still alive, else None."""
        if self._viewer_ref is None:
            return None
        return self._viewer_ref()

    def set_viewer(self, viewer):
        """Set a weak reference to the owning viewer."""
        self._viewer_ref = weakref(viewer) if viewer is not None else None

    # ----- Convenience property accessors that match Common naming -----

    def update_canvas(self, canvas):
        """Update the canvas reference."""
        self.canvas = canvas

    def update_fig(self, fig):
        """Update the figure reference."""
        self.fig = fig

    def update_ax(self, ax):
        """Update the axes reference."""
        self.ax = ax

    def update_overlay_ax(self, overlay_ax):
        """Update the overlay axes reference."""
        self.overlay_ax = overlay_ax

    def update_im(self, im):
        """Update the image reference."""
        self.im = im

    def update_background(self, background):
        """Update the cached background for blitting."""
        self._background = background

    def update_lines(self, hline, vline, cpoint=None):
        """Update crosshair line references."""
        self.hline = hline
        self.vline = vline
        self.cpoint = cpoint

    def update_colorbar(self, colorbar):
        """Update the colorbar reference."""
        self.colorbar = colorbar

    def update_cax(self, cax):
        """Update the colorbar axes reference."""
        self.cax = cax

    def update_poslabel(self, label):
        """Update the position label reference."""
        self.plabel = label

    def update_chlabel(self, label):
        """Update the channel label reference."""
        self.chlabel = label

    def update_hpbw(self, hpbw):
        """Update the HPBW ellipse reference."""
        self.hpbw = hpbw

    def update_slider(self, slider):
        """Update the slider reference."""
        self.slider = slider

    def update_chval_box(self, textbox):
        """Update the channel value textbox reference."""
        self.chval_box = textbox

    def update_xrange_input(self, xmin_input, xmax_input):
        """Update X-range input field references."""
        self.xmin_input = xmin_input
        self.xmax_input = xmax_input

    def update_yrange_input(self, ymin_input, ymax_input):
        """Update Y-range input field references."""
        self.ymin_input = ymin_input
        self.ymax_input = ymax_input

    def update_zrange_input(self, zmin_input, zmax_input):
        """Update Z-range input field references."""
        self.zmin_input = zmin_input
        self.zmax_input = zmax_input

    def update_ax_units(self, xunit, yunit):
        """Update axis unit metadata."""
        self.ax_xunit = xunit
        self.ax_yunit = yunit

    def update_ax_types(self, xtype, ytype):
        """Update axis type metadata."""
        self.ax_xtype = xtype
        self.ax_ytype = ytype

    def update_ax_coord(self, ax_coord):
        """Update the axis coordinate reference list."""
        self.ax_coord = ax_coord

    def copy_overlay_background(self):
        """
        Capture the overlay background for blitting.
        Returns the captured background or None if not possible.
        """
        if self.canvas is None or self.overlay_ax is None:
            return None

        viewer = self.viewer
        hidden_patches = []
        hidden_markers = []
        hidden_artists = []
        region_manager = None
        marker_manager = None

        if viewer is not None:
            region_manager = getattr(viewer, 'region_manager', None)
            marker_manager = getattr(viewer, 'marker_manager', None)
            if region_manager is not None:
                hidden_patches = region_manager.prepare_for_background_capture()
            if marker_manager is not None:
                hidden_markers = marker_manager.prepare_for_background_capture(self.plane)

        # Hide cursor/label artists so they are not baked into the background.
        for artist in (self.hline, self.vline, self.cpoint, self.chlabel):
            if artist is None:
                continue
            try:
                if artist.get_visible():
                    artist.set_visible(False)
                    hidden_artists.append(artist)
            except Exception:
                continue
        hpbw = getattr(self, 'hpbw', None)
        hpbw_artist = getattr(hpbw, 'ellipse', None) if hpbw is not None else None
        if hpbw_artist is not None:
            try:
                if hpbw_artist.get_visible():
                    hpbw_artist.set_visible(False)
                    hidden_artists.append(hpbw_artist)
            except Exception:
                pass

        background = self.canvas.copy_from_bbox(self.overlay_ax.bbox)

        if region_manager is not None:
            region_manager.restore_after_background_capture(hidden_patches)
            region_manager.draw_regions_for_blit()
        if marker_manager is not None:
            if hidden_markers:
                marker_manager.restore_after_background_capture(hidden_markers)
            marker_manager.draw_markers_for_blit()
            if self.plane == 'xy':
                marker_manager.redraw_planes(['xz', 'zy'])

        for artist in hidden_artists:
            try:
                artist.set_visible(True)
            except Exception:
                pass

        return background
